#include <stdarg.h>
#include <sys/select.h>
#include <stdint.h>
#include <stdlib.h>
#include "gfx.h"
#include "../lib/proto.h"
#define FIKE_IMP_CLIENT
#include "../lib/fike.h"
#include "../lib/fike.c"


int length(int dx, int dy) { return dx*dx+dy*dy; }
#define CLICK_THRESHOLD 100

static int dpi;
typedef struct { uint32_t host; int port; char *name; } host_t;
static FIKE_SERVER *server;
static host_t *hostinfo;

void fike_add_fd(int fd) 
{
  gfx_monitor_file(fd, 213, 1,0,1);
}
void fike_remove_fd(int fd) 
{
  gfx_stop_monitor_file(fd,-1);
}

void fike_found_server
       (char *host, uint32_t addr, int port, int type, char *name)
{
  host_t *H;
  if (type!=STYPE_MOUSE) return ;
  H= malloc(sizeof(*H));
  H->name= strdup(name);
  H->port= port;
  H->host= addr;
  gfx_insert(WssLIST,-1, name, H,NULL);
}

static void process_ready_fike_fd(int fd)
{
  fd_set S, X;
  FD_ZERO(&S);
  FD_ZERO(&X);
  FD_SET(fd, &S);
  fike_process_select(&S, &X);
}

static void process_except_fike_fd(int fd)
{
  fd_set S, X;
  FD_ZERO(&S);
  FD_ZERO(&X);
  FD_SET(fd, &X);
  fike_process_select(&S, &X);
}

static void select_host()
{
  gfx_event_t event;
  host_t *HH;
  HH= NULL;
  int q;
  q= 0;
  gfx_SCREEN_server_select();
  fike_start_query(NULL, "*.mouse");
  while(!q) if (gfx_event(NULL,&event)) switch(event.type)
  {
    // FIXME: we should put this in the library itself.
  case GFXEV_READY_READ: process_ready_fike_fd(event.wide); break;
  case GFXEV_READY_EXCEPT: process_except_fike_fd(event.wide); break;
  case GFXEV_QUIT:
  case WssCANCEL: q= 2;break;
  case WssOK: if (HH) q= 1; else q= 2; break;
  case WssLIST: HH= event.usr; break;
  }
  fike_stop_query();
  gfx_leave_screen();
  if (q==2 || !HH) goto done;
  if (hostinfo && hostinfo->host==HH->host) goto done;
  if (server) fike_close_server(server);
  server= NULL;
  if ((server=fike_connect(HH->host, HH->port)))
     hostinfo= HH;
done:
  if (hostinfo) gfx_set_text(Whost, hostinfo->name);
           else gfx_set_text(Whost, "");
}

static void send_command(int cmd, int a, int b)
{
  if (!server) return ;
  fike_command(server, 0, cmd, "ww", a,b);
}

static void toggle_button(int widget, int button)
{
  int state;
  state= gfx_get_state(widget);
  send_command(CMD_BUTTON, button, state);
}

static void toggle_key(int widget, int key)
{
  int state;
  state= gfx_get_state(widget);
  send_command(CMD_CONTROL_KEY, key, state);
}

static void click_button(int button)
{
  send_command(CMD_BUTTON, button, 1);
  usleep(10000);
  send_command(CMD_BUTTON, button, 0);
}

static void report_delta(int dx,int dy)
{
  if (dpi) { dx = dx*96/dpi; dy = dy*96/dpi; }
  send_command(CMD_RELMOVE, dx, dy);
}

static void main_event_loop()
{
  gfx_event_t event;
  int mx, my, pc, px, py;
  while(1) if (gfx_event(NULL,&event)) switch(event.type)
  {
    case GFXEV_QUIT: return ;
    case Wconnect: select_host(); break;
    case WtogSHIFT: toggle_key(event.type, KK_SHIFT); break;
    case WtogCTRL: toggle_key(event.type, KK_CTRL); break;
    case WtogLEFT: toggle_button(event.type,0); break;
    case WtogMIDDLE: toggle_button(event.type,2); break;
    case WtogRIGHT: toggle_button(event.type,1); break;
    case WbutLEFT: click_button(0); break;
    case WbutRIGHT: click_button(1); break;
    case WbutMIDDLE: click_button(2); break;
    case GFXEV(Wpad, BUTTON_PRESS):
       gfx_canvas_begin_report_mouse_position(Wpad);
       px= mx= event.x; py= my= event.y;
       pc= 0;
       break;
    case GFXEV(Wpad, BUTTON_RELEASE):
       if (length(event.x-px,event.y-py)<CLICK_THRESHOLD)
          click_button(0);
       else
          report_delta(event.x-mx, event.y-my);
       gfx_canvas_end_report_mouse_position(Wpad);
       break;
    case GFXEV(Wpad, POINTER_MOVE):
       pc++;
       report_delta(event.x-mx, event.y-my);
       // printf("hooho\n");
       mx= event.x; my= event.y;
       break;
  }
}


static void main_loop()
{
  unsigned int D, *DD;
  gfx_SCREEN_0();
  dpi= gfx_get_dpi();
  select_host();
  gfx_canvas_get_buffer(Wpad, &D, &D, &DD);
  main_event_loop();
}


#ifdef GFX_X11
int main()
{
  gfx_init(NULL, 400, 600, "Remote Mouse");
  main_loop();
}
#endif

#ifdef GFX_ANDROID
void android_main(unsigned char *state, int len)
{
  main_loop();
}
#endif

