#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <sys/select.h>
#include <fcntl.h>

#define FIKE_IMP_SERVER
#include "../lib/proto.h"
#include "../lib/fike.h"
#include "../lib/fike.c"

static int ifd= -1;

void* fike_client_connected(char *host, int port, int fd)
{
  static int A;
  return &A;
}
void fike_client_disconnected(char *host, int port, void *udata)
{
}

void die(char *call)
{
  fprintf(stderr,"%s failed: %s\n", call, strerror(errno));
  exit(1);
}

struct
{
  char *name; int v1, v2;
} ctls[]=
{
  { "Enable synchronization", UI_SET_EVBIT, EV_SYN },
  { "Enable key events",      UI_SET_EVBIT, EV_KEY },
  { "Enable button 0",        UI_SET_KEYBIT, BTN_LEFT },
  { "Enable button 1",        UI_SET_KEYBIT, BTN_RIGHT },
  { "Enable button 2",        UI_SET_KEYBIT, BTN_MIDDLE },
  { "Enable CTRL",            UI_SET_KEYBIT, KEY_LEFTCTRL },
  { "Enable SHIFT",           UI_SET_KEYBIT, KEY_LEFTSHIFT },
  { "Enable pointer events",  UI_SET_EVBIT, EV_REL },
  { "Enable X axis",          UI_SET_RELBIT, REL_X },
  { "Enable Y axis",          UI_SET_RELBIT, REL_Y },
  { NULL },
};

static void isetup()
{
   int i;
   ifd= open("/dev/uinput", O_WRONLY | O_NONBLOCK);
   if (ifd<0) die("open");

   for(i=0;ctls[i].name;i++)
     if (ioctl(ifd, ctls[i].v1, ctls[i].v2))
        die(ctls[i].name);

#if 0  // new interface, I have old kernel
   struct uinput_setup usetup;
   memset(&usetup, 0, sizeof(usetup));
   usetup.id.bustype = BUS_USB;
   usetup.id.vendor = 0x1; /* sample vendor */
   usetup.id.product = 0x1; /* sample product */
   usetup.id.version= 1;
   strcpy(usetup.name, "uinput-sle0");
   ioctl(ifd, UI_DEV_SETUP, &usetup);
#else
   struct uinput_user_dev dev;
   memset(&dev,0,sizeof(dev));
   dev.id.bustype= BUS_USB;
   dev.id.vendor= 1;
   dev.id.product= 1;
   dev.id.version= 1;
   strncpy(dev.name, "uinput-sle0", UINPUT_MAX_NAME_SIZE);
   if (write(ifd, &dev, sizeof(dev))<0) die("configure");
#endif

   if (ioctl(ifd, UI_DEV_CREATE)) die("create");
   return ;

}

static void ifinish()
{
  ioctl(ifd, UI_DEV_DESTROY);
  close(ifd);
  ifd= -1;
}

static void ievent(int type, int code, int value)
{
  struct input_event ev;
  memset(&ev, 0 , sizeof(ev));
  ev.type= type;
  ev.code= code;
  ev.value= value;
//  ev.time.tv_sec= 0;
//  ev.time.tv_usec= 0;
   gettimeofday(&ev.time, NULL);
  if (write(ifd, &ev, sizeof(ev))<0) 
     fprintf(stderr,"write: %s\n",strerror(errno));;
  // if (type==EV_SYN) usleep(5*1000);
}

#ifndef TESTING
static void move_mouse(int dx, int dy)
{
  ievent( EV_REL, REL_X, dx);
  ievent( EV_REL, REL_Y, dy);
  ievent( EV_SYN, SYN_REPORT, 0);
}

static void mouse_button(int button, int state)
{
  switch(button) 
  {
   case 0: button= BTN_LEFT; break;
   case 1: button= BTN_RIGHT; break;
   case 2: button= BTN_MIDDLE; break;
   default: return ;
  }
  ievent( EV_KEY, button, state );
  ievent( EV_SYN, SYN_REPORT, 0);
}

static void control_key_event(int key,int state)
{
  switch(key)
  {
  case KK_CTRL: key= KEY_LEFTCTRL; break;
  case KK_SHIFT: key= KEY_LEFTSHIFT; break;
  default: return ;
  }
  ievent( EV_KEY, key, state );
  ievent(EV_SYN, SYN_REPORT, 0);
}
#else
static void move_mouse(int dx, int dy)
{
  printf("move: %d, %d\n", dx, dy);
}
static void mouse_button(int button, int state)
{
  printf("button: %d= %d\n", button, state);
}

static void control_key_event(int key,int state)
{
  printf("control key: %d= %d\n", key, state);
}
#endif

void fike_client_message(FIKE_CLIENT *cc,void *usr, int cmd)
{
  short dx, dy, button, state, key;
  switch(cmd)
  {
  case CMD_RELMOVE: 
    fike_command_params(cc, "ww", &dx,&dy); 
    move_mouse(dx, dy);
    break;
  case CMD_BUTTON:
    fike_command_params(cc, "ww", &button, &state);
    mouse_button(button, state);
    break;
  case CMD_CONTROL_KEY:
    fike_command_params(cc,"ww",&key,&state);
    control_key_event(key,state);
    break;
  }
}

void fike_add_fd(int fd) { }
void fike_remove_fd(int fd) { }

int main(int argc,char **argv)
{
#if 0
  isetup(); sleep(2);
  mouse_button(0,1);
  usleep(10000);
  mouse_button(1,0);

  move_mouse(-100,-100);
  ifinish();
#else


  if (argc!=2)  exit(fprintf(stderr,"usage: %s name\n",argv[0])); 
  isetup();
  if (ifd<0) exit(fprintf(stderr,"can't do uinput\n"));
  fike_publish(NULL, argv[1], STYPE_MOUSE, STYPE_MOUSE);
  while(1)
  {
   fd_set rset, xset;
   int max;
   int R;
   max= fike_get_select(&rset,&xset);
   R= select(max+1,&rset, NULL, &xset, NULL);
   if (R<0) { if (errno!=EINTR) usleep(10000); usleep(10000); }
   fike_process_select(&rset, &xset);
  }
  ifinish();
#endif
}
